// Copyright 2019 Kina Wakasa
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

class NullKeyError extends Error {}

class EmptyKeyError extends Error {}

class ReadOp {
  final String key;

  ReadOp(this.key) {
    if (key == null) {
      throw NullKeyError();
    }
    if (key.isEmpty) {
      throw EmptyKeyError();
    }
  }
}

class WriteOp {
  final String key;
  final String value;

  WriteOp(this.key, this.value) {
    if (key == null) {
      throw NullKeyError();
    }
    if (key.isEmpty) {
      throw EmptyKeyError();
    }
  }
}

class ReadBatch {
  final List<ReadOp> ops = List<ReadOp>();

  void get(String key) => ops.add(ReadOp(key));
}

class WriteBatch {
  final ops = List<WriteOp>();

  void put(String key, String value) => ops.add(WriteOp(key, value));

  void delete(String key) => ops.add(WriteOp(key, null));
}

class NullLoKVNameError extends Error {}

class EmptyLoKVNameError extends Error {}

class ClosedLoKVError extends Error {}

enum LoKVStatus { closed, open }

abstract class LoKV {
  final String name;
  LoKVStatus _status;

  LoKVStatus get status => _status;

  LoKV(this.name) : _status = LoKVStatus.closed;

  factory LoKV.create(String name) {
    if (name == null) {
      throw NullLoKVNameError();
    }
    if (name.isEmpty) {
      throw EmptyLoKVNameError();
    }
    return VolatileLoKV(name);
  }

  Future<void> _open();

  Future<void> open() async {
    if (_status == LoKVStatus.closed) {
      await _open();
      _status = LoKVStatus.open;
    }
  }

  Future<void> _close();

  Future<void> close() async {
    if (_status == LoKVStatus.open) {
      await _close();
      _status = LoKVStatus.closed;
    }
  }

  Future<void> _drop();

  Future<void> drop() async {
    await close();
    await _drop();
  }

  Future<List<String>> _read(List<ReadOp> ops);

  Future<void> _write(List<WriteOp> ops);

  Future<List<String>> read(ReadBatch batch) {
    if (_status == LoKVStatus.closed) {
      throw ClosedLoKVError();
    }
    return _read(batch.ops);
  }

  Future<void> write(WriteBatch batch) {
    if (_status == LoKVStatus.closed) {
      throw ClosedLoKVError();
    }
    return _write(batch.ops);
  }

  Future<String> get(String key) async =>
      (await read(ReadBatch()..get(key)))[0];

  Future<void> put(String key, String value) async =>
      write(WriteBatch()..put(key, value));

  Future<void> delete(String key) async => write(WriteBatch()..delete(key));
}

class VolatileLoKV extends LoKV {
  Map<String, String> _db;

  VolatileLoKV(String name) : super(name);

  @override
  Future<void> _open() async {
    if (_db == null) {
      _db = Map<String, String>();
    }
  }

  @override
  Future<void> _close() async {}

  @override
  Future<void> _drop() async {
    _db = null;
  }

  @override
  Future<List<String>> _read(List<ReadOp> ops) async {
    var values = List<String>(ops.length);
    for (var i = 0; i < ops.length; ++i) {
      values[i] = _db[ops[i].key];
    }
    return values;
  }

  @override
  Future<void> _write(List<WriteOp> ops) async {
    for (var op in ops) {
      if (op.value == null) {
        _db.remove(op.key);
      } else {
        _db[op.key] = op.value;
      }
    }
  }
}
